"""
Abstract Continuous Function
============================

Defines the abstract class implementing support for abstract continuous
function:

-   :class:`colour.continuous.AbstractContinuousFunction`.
"""

from __future__ import annotations

import numpy as np
from abc import ABC, abstractmethod
from copy import deepcopy

from colour.hints import (
    ArrayLike,
    Any,
    Boolean,
    Callable,
    Dict,
    DTypeFloating,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Integer,
    Literal,
    NDArray,
    Number,
    Optional,
    Type,
    TypeExtrapolator,
    TypeInterpolator,
    Union,
)
from colour.utilities import (
    as_float,
    attest,
    closest,
    is_uniform,
    is_string,
    optional,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "AbstractContinuousFunction",
]


class AbstractContinuousFunction(ABC):
    """
    Define the base class for abstract continuous function.

    This is an :class:`ABCMeta` abstract class that must be inherited by
    sub-classes.

    The sub-classes are expected to implement the
    :meth:`colour.continuous.AbstractContinuousFunction.function` method so
    that evaluating the function for any independent domain
    variable :math:`x \\in\\mathbb{R}` returns a corresponding range variable
    :math:`y \\in\\mathbb{R}`. A conventional implementation adopts an
    interpolating function encapsulated inside an extrapolating function.
    The resulting function independent domain, stored as discrete values in the
    :attr:`colour.continuous.AbstractContinuousFunction.domain` attribute
    corresponds with the function dependent and already known range stored in
    the :attr:`colour.continuous.AbstractContinuousFunction.range` property.

    Parameters
    ----------
    name
        Continuous function name.

    Attributes
    ----------
    -   :attr:`~colour.continuous.AbstractContinuousFunction.name`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.dtype`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.domain`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.range`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.interpolator`
    -   :attr:`~colour.continuous.\
AbstractContinuousFunction.interpolator_kwargs`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.extrapolator`
    -   :attr:`~colour.continuous.\
AbstractContinuousFunction.extrapolator_kwargs`
    -   :attr:`~colour.continuous.AbstractContinuousFunction.function`

    Methods
    -------
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__init__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__str__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__repr__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__hash__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__getitem__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__setitem__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__contains__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__len__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__eq__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__ne__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__iadd__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__add__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__isub__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__sub__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__imul__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__mul__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__idiv__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__div__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__ipow__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.__pow__`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.\
arithmetical_operation`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.fill_nan`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.domain_distance`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.is_uniform`
    -   :meth:`~colour.continuous.AbstractContinuousFunction.copy`
    """

    def __init__(self, name: Optional[str] = None):
        self._name: str = f"{self.__class__.__name__} ({id(self)})"
        self.name = optional(name, self._name)

    @property
    def name(self) -> str:
        """
        Getter and setter property for the abstract continuous function name.

        Parameters
        ----------
        value
            Value to set the abstract continuous function name with.

        Returns
        -------
        :class:`str`
            Abstract continuous function name.
        """

        return self._name

    @name.setter
    def name(self, value: str):
        """Setter for the **self.name** property."""

        attest(
            is_string(value),
            f'"name" property: "{value}" type is not "str"!',
        )

        self._name = value

    # TODO: Remove pragma when https://github.com/python/mypy/issues/4165 is
    # resolved.
    @property  # type: ignore[misc]
    @abstractmethod
    def dtype(self) -> Type[DTypeFloating]:
        """
        Getter and setter property for the abstract continuous function dtype,
        must be reimplemented by sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function dtype with.

        Returns
        -------
        Type[DTypeFloating]
            Abstract continuous function dtype.
        """

        ...  # pragma: no cover

    @dtype.setter  # type: ignore[misc]
    @abstractmethod
    def dtype(self, value: Type[DTypeFloating]):
        """
        Setter for the **self.dtype** property, must be reimplemented by
        sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def domain(self) -> NDArray:
        """
        Getter and setter property for the abstract continuous function
        independent domain variable :math:`x`, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function independent domain
            variable :math:`x` with.

        Returns
        -------
        :class:`numpy.ndarray`
            Abstract continuous function independent domain variable :math:`x`.
        """

        ...  # pragma: no cover

    @domain.setter  # type: ignore[misc]
    @abstractmethod
    def domain(self, value: ArrayLike):
        """
        Setter for the **self.domain** property, must be reimplemented by
        sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def range(self) -> NDArray:
        """
        Getter and setter property for the abstract continuous function
        corresponding range variable :math:`y`, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function corresponding range
            variable :math:`y` with.

        Returns
        -------
        :class:`numpy.ndarray`
            Abstract continuous function corresponding range variable
            :math:`y`.
        """

        ...  # pragma: no cover

    @range.setter  # type: ignore[misc]
    @abstractmethod
    def range(self, value: ArrayLike):
        """
        Setter for the **self.range** property, must be reimplemented by
        sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def interpolator(self) -> Type[TypeInterpolator]:
        """
        Getter and setter property for the abstract continuous function
        interpolator type, must be reimplemented by sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function interpolator type
            with.

        Returns
        -------
        Type[TypeInterpolator]
            Abstract continuous function interpolator type.
        """

        ...  # pragma: no cover

    @interpolator.setter  # type: ignore[misc]
    @abstractmethod
    def interpolator(self, value: Type[TypeInterpolator]):
        """
        Setter for the **self.interpolator** property, must be reimplemented by
        sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def interpolator_kwargs(self) -> Dict:
        """
        Getter and setter property for the abstract continuous function
        interpolator instantiation time arguments, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function interpolator
            instantiation time arguments to.

        Returns
        -------
        :class:`dict`
            Abstract continuous function interpolator instantiation time
            arguments.
        """

        ...  # pragma: no cover

    @interpolator_kwargs.setter  # type: ignore[misc]
    @abstractmethod
    def interpolator_kwargs(self, value: dict):
        """
        Setter for the **self.interpolator_kwargs** property, must be
        reimplemented by sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def extrapolator(self) -> Type[TypeExtrapolator]:
        """
        Getter and setter property for the abstract continuous function
        extrapolator type, must be reimplemented by sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function extrapolator type
            with.

        Returns
        -------
        Type[TypeExtrapolator]
            Abstract continuous function extrapolator type.
        """

        ...  # pragma: no cover

    @extrapolator.setter  # type: ignore[misc]
    @abstractmethod
    def extrapolator(self, value: Type[TypeExtrapolator]):
        """
        Setter for the **self.extrapolator** property, must be reimplemented by
        sub-classes.
        """

        ...  # pragma: no cover

    @property  # type: ignore[misc]
    @abstractmethod
    def extrapolator_kwargs(self) -> Dict:
        """
        Getter and setter property for the abstract continuous function
        extrapolator instantiation time arguments, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        value
            Value to set the abstract continuous function extrapolator
            instantiation time arguments to.

        Returns
        -------
        :class:`dict`
            Abstract continuous function extrapolator instantiation time
            arguments.
        """

        ...  # pragma: no cover

    @extrapolator_kwargs.setter  # type: ignore[misc]
    @abstractmethod
    def extrapolator_kwargs(self, value: dict):
        """
        Setter for the **self.extrapolator_kwargs** property, must be
        reimplemented by sub-classes.
        """

        ...  # pragma: no cover

    @property
    @abstractmethod
    def function(self) -> Callable:
        """
        Getter property for the abstract continuous function callable, must be
        reimplemented by sub-classes.

        Returns
        -------
        Callable
            Abstract continuous function callable.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __str__(self) -> str:
        """
        Return a formatted string representation of the abstract continuous
        function, must be reimplemented by sub-classes.

        Returns
        -------
        :class:`str`
            Formatted string representation.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __repr__(self) -> str:
        """
        Return an evaluable string representation of the abstract continuous
        function, must be reimplemented by sub-classes.

        Returns
        -------
        :class:`str`
            Evaluable string representation.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __hash__(self) -> Integer:
        """
        Return the abstract continuous function hash.

        Returns
        -------
        :class:`numpy.integer`
            Object hash.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __getitem__(
        self, x: Union[FloatingOrArrayLike, slice]
    ) -> FloatingOrNDArray:
        """
        Return the corresponding range variable :math:`y` for independent
        domain variable :math:`x`, must be reimplemented by sub-classes.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Variable :math:`y` range value.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __setitem__(
        self, x: Union[FloatingOrArrayLike, slice], y: FloatingOrArrayLike
    ):
        """
        Set the corresponding range variable :math:`y` for independent domain
        variable :math:`x`, must be reimplemented by sub-classes.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.
        y
            Corresponding range variable :math:`y`.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __contains__(self, x: Union[FloatingOrArrayLike, slice]) -> bool:
        """
        Return whether the abstract continuous function contains given
        independent domain variable :math:`x`, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.

        Returns
        -------
        :class:`bool`
            Whether :math:`x` domain value is contained.
        """

        ...  # pragma: no cover

    def __len__(self) -> Integer:
        """
        Return the abstract continuous function independent domain :math:`x`
        variable elements count.


        Returns
        -------
        :class:`numpy.integer`
            Independent domain variable :math:`x` elements count.
        """

        return len(self.domain)

    @abstractmethod
    def __eq__(self, other: Any) -> bool:
        """
        Return whether the abstract continuous function is equal to given
        other object, must be reimplemented by sub-classes.

        Parameters
        ----------
        other
            Object to test whether it is equal to the abstract continuous
            function.

        Returns
        -------
        :class:`bool`
            Whether given object is equal to the abstract continuous function.
        """

        ...  # pragma: no cover

    @abstractmethod
    def __ne__(self, other: Any) -> bool:
        """
        Return whether the abstract continuous function is not equal to given
        other object, must be reimplemented by sub-classes.

        Parameters
        ----------
        other
            Object to test whether it is not equal to the abstract continuous
            function.

        Returns
        -------
        :class:`bool`
            Whether given object is not equal to the abstract continuous
            function.
        """

        ...  # pragma: no cover

    def __add__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for addition.

        Parameters
        ----------
        a
            Variable :math:`a` to add.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Variable added abstract continuous function.
        """

        return self.arithmetical_operation(a, "+")

    def __iadd__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for in-place addition.

        Parameters
        ----------
        a
            Variable :math:`a` to add in-place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            In-place variable added abstract continuous function.
        """

        return self.arithmetical_operation(a, "+", True)

    def __sub__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for subtraction.

        Parameters
        ----------
        a
            Variable :math:`a` to subtract.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Variable subtracted abstract continuous function.
        """

        return self.arithmetical_operation(a, "-")

    def __isub__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for in-place subtraction.

        Parameters
        ----------
        a
            Variable :math:`a` to subtract in-place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            In-place variable subtracted abstract continuous function.
        """

        return self.arithmetical_operation(a, "-", True)

    def __mul__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for multiplication.

        Parameters
        ----------
        a
            Variable :math:`a` to multiply by.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Variable multiplied abstract continuous function.
        """

        return self.arithmetical_operation(a, "*")

    def __imul__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for in-place multiplication.

        Parameters
        ----------
        a
            Variable :math:`a` to multiply by in-place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            In-place variable multiplied abstract continuous function.
        """

        return self.arithmetical_operation(a, "*", True)

    def __div__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for division.

        Parameters
        ----------
        a
            Variable :math:`a` to divide by.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Variable divided abstract continuous function.
        """

        return self.arithmetical_operation(a, "/")

    def __idiv__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for in-place division.

        Parameters
        ----------
        a
            Variable :math:`a` to divide by in-place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            In-place variable divided abstract continuous function.
        """

        return self.arithmetical_operation(a, "/", True)

    __itruediv__ = __idiv__
    __truediv__ = __div__

    def __pow__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for exponentiation.

        Parameters
        ----------
        a
            Variable :math:`a` to exponentiate by.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Variable exponentiated abstract continuous function.
        """

        return self.arithmetical_operation(a, "**")

    def __ipow__(
        self, a: Union[FloatingOrArrayLike, AbstractContinuousFunction]
    ) -> AbstractContinuousFunction:
        """
        Implement support for in-place exponentiation.

        Parameters
        ----------
        a
            Variable :math:`a` to exponentiate by in-place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            In-place variable exponentiated abstract continuous function.
        """

        return self.arithmetical_operation(a, "**", True)

    @abstractmethod
    def arithmetical_operation(
        self,
        a: Union[FloatingOrArrayLike, AbstractContinuousFunction],
        operation: Literal["+", "-", "*", "/", "**"],
        in_place: Boolean = False,
    ) -> AbstractContinuousFunction:
        """
        Perform given arithmetical operation with operand :math:`a`, the
        operation can be either performed on a copy or in-place, must be
        reimplemented by sub-classes.

        Parameters
        ----------
        a
            Operand :math:`a`.
        operation
            Operation to perform.
        in_place
            Operation happens in place.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Abstract continuous function.
        """

        ...  # pragma: no cover

    @abstractmethod
    def fill_nan(
        self,
        method: Union[
            Literal["Constant", "Interpolation"], str
        ] = "Interpolation",
        default: Number = 0,
    ) -> AbstractContinuousFunction:
        """
        Fill NaNs in independent domain variable :math:`x` and corresponding
        range variable :math:`y` using given method, must be reimplemented by
        sub-classes.

        Parameters
        ----------
        method
            *Interpolation* method linearly interpolates through the NaNs,
            *Constant* method replaces NaNs with ``default``.
        default
            Value to use with the *Constant* method.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            NaNs filled abstract continuous function.
        """

        ...  # pragma: no cover

    def domain_distance(self, a: FloatingOrArrayLike) -> FloatingOrNDArray:
        """
        Return the euclidean distance between given array and independent
        domain :math:`x` closest element.

        Parameters
        ----------
        a
            Variable :math:`a` to compute the euclidean distance with
            independent domain variable :math:`x`.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Euclidean distance between independent domain variable :math:`x`
            and given variable :math:`a`.
        """

        n = closest(self.domain, a)

        return as_float(np.abs(a - n))

    def is_uniform(self) -> Boolean:
        """
        Return if independent domain variable :math:`x` is uniform.

        Returns
        -------
        :class:`bool`
            Is independent domain variable :math:`x` uniform.
        """

        return is_uniform(self.domain)

    def copy(self) -> AbstractContinuousFunction:
        """
        Return a copy of the sub-class instance.

        Returns
        -------
        :class:`colour.continuous.AbstractContinuousFunction`
            Abstract continuous function copy.
        """

        return deepcopy(self)



"""
Cylindrical & Spherical Colour Models
=====================================

Defines various cylindrical and spherical colour models:

-   :func:`colour.RGB_to_HSV`
-   :func:`colour.HSV_to_RGB`
-   :func:`colour.RGB_to_HSL`
-   :func:`colour.HSL_to_RGB`
-   :func:`colour.RGB_to_HCL`
-   :func:`colour.HCL_to_RGB`

These colour models trade off perceptual relevance for computation speed.
They should not be used in the colour science domain although they are useful
for image analysis and provide end user software colour selection tools.

They are provided for convenience and completeness.

References
----------
-   :cite:`EasyRGBj` : EasyRGB. (n.d.). RGB --> HSV. Retrieved May 18, 2014,
    from http://www.easyrgb.com/index.php?X=MATH&H=20#text20
-   :cite:`EasyRGBk` : EasyRGB. (n.d.). HSL --> RGB. Retrieved May 18, 2014,
    from http://www.easyrgb.com/index.php?X=MATH&H=19#text19
-   :cite:`EasyRGBl` : EasyRGB. (n.d.). RGB --> HSL. Retrieved May 18, 2014,
    from http://www.easyrgb.com/index.php?X=MATH&H=18#text18
-   :cite:`EasyRGBn` : EasyRGB. (n.d.). HSV --> RGB. Retrieved May 18, 2014,
    from http://www.easyrgb.com/index.php?X=MATH&H=21#text21
-   :cite:`Smith1978b` : Smith, A. R. (1978). Color gamut transform pairs.
    Proceedings of the 5th Annual Conference on Computer Graphics and
    Interactive Techniques - SIGGRAPH "78, 12-19. doi:10.1145/800248.807361
-   :cite:`Wikipedia2003` : Wikipedia. (2003). HSL and HSV. Retrieved
    September 10, 2014, from http://en.wikipedia.org/wiki/HSL_and_HSV
-   :cite:`Sarifuddin2005` : Sarifuddin, M., & Missaoui, R. (2005). A New
    Perceptually Uniform Color Space with Associated Color Similarity Measure
    for ContentBased Image and Video Retrieval.
-   :cite:`Wikipedia2015` : Wikipedia. (2015). HCL color space. Retrieved
    April 4, 2021, from https://en.wikipedia.org/wiki/HCL_color_space
"""

from __future__ import annotations

import numpy as np

from colour.hints import ArrayLike, Floating, NDArray
from colour.utilities import (
    as_float_array,
    from_range_1,
    to_domain_1,
    tsplit,
    tstack,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "RGB_to_HSV",
    "HSV_to_RGB",
    "RGB_to_HSL",
    "HSL_to_RGB",
    "RGB_to_HCL",
    "HCL_to_RGB",
]


def RGB_to_HSV(RGB: ArrayLike) -> NDArray:
    """
    Convert from *RGB* colourspace to *HSV* colourspace.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *HSV* array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HSV``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`EasyRGBj`, :cite:`Smith1978b`, :cite:`Wikipedia2003`

    Examples
    --------
    >>> RGB = np.array([0.45620519, 0.03081071, 0.04091952])
    >>> RGB_to_HSV(RGB)  # doctest: +ELLIPSIS
    array([ 0.9960394...,  0.9324630...,  0.4562051...])
    """

    RGB = to_domain_1(RGB)

    maximum = np.amax(RGB, -1)
    delta = np.ptp(RGB, -1)

    V = maximum

    R, G, B = tsplit(RGB)

    S = as_float_array(delta / maximum)
    S[np.asarray(delta == 0)] = 0

    delta_R = (((maximum - R) / 6) + (delta / 2)) / delta
    delta_G = (((maximum - G) / 6) + (delta / 2)) / delta
    delta_B = (((maximum - B) / 6) + (delta / 2)) / delta

    H = delta_B - delta_G
    H = np.where(G == maximum, (1 / 3) + delta_R - delta_B, H)
    H = np.where(B == maximum, (2 / 3) + delta_G - delta_R, H)
    H[np.asarray(H < 0)] += 1
    H[np.asarray(H > 1)] -= 1
    H[np.asarray(delta == 0)] = 0

    HSV = tstack([H, S, V])

    return from_range_1(HSV)


def HSV_to_RGB(HSV: ArrayLike) -> NDArray:
    """
    Convert from *HSV* colourspace to *RGB* colourspace.

    Parameters
    ----------
    HSV
        *HSV* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HSV``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`EasyRGBn`, :cite:`Smith1978b`, :cite:`Wikipedia2003`

    Examples
    --------
    >>> HSV = np.array([0.99603944, 0.93246304, 0.45620519])
    >>> HSV_to_RGB(HSV)  # doctest: +ELLIPSIS
    array([ 0.4562051...,  0.0308107...,  0.0409195...])
    """

    H, S, V = tsplit(to_domain_1(HSV))

    h = as_float_array(H * 6)
    h[np.asarray(h == 6)] = 0

    i = np.floor(h)
    j = V * (1 - S)
    k = V * (1 - S * (h - i))
    l = V * (1 - S * (1 - (h - i)))  # noqa

    i = tstack([i, i, i]).astype(np.uint8)

    RGB = np.choose(
        i,
        [
            tstack([V, l, j]),
            tstack([k, V, j]),
            tstack([j, V, l]),
            tstack([j, k, V]),
            tstack([l, j, V]),
            tstack([V, j, k]),
        ],
        mode="clip",
    )

    return from_range_1(RGB)


def RGB_to_HSL(RGB: ArrayLike) -> NDArray:
    """
    Convert from *RGB* colourspace to *HSL* colourspace.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *HSL* array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HSL``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`EasyRGBl`, :cite:`Smith1978b`, :cite:`Wikipedia2003`

    Examples
    --------
    >>> RGB = np.array([0.45620519, 0.03081071, 0.04091952])
    >>> RGB_to_HSL(RGB)  # doctest: +ELLIPSIS
    array([ 0.9960394...,  0.8734714...,  0.2435079...])
    """

    RGB = to_domain_1(RGB)

    minimum = np.amin(RGB, -1)
    maximum = np.amax(RGB, -1)
    delta = np.ptp(RGB, -1)

    R, G, B = tsplit(RGB)

    L = (maximum + minimum) / 2

    S = np.where(
        L < 0.5,
        delta / (maximum + minimum),
        delta / (2 - maximum - minimum),
    )
    S[np.asarray(delta == 0)] = 0

    delta_R = (((maximum - R) / 6) + (delta / 2)) / delta
    delta_G = (((maximum - G) / 6) + (delta / 2)) / delta
    delta_B = (((maximum - B) / 6) + (delta / 2)) / delta

    H = delta_B - delta_G
    H = np.where(G == maximum, (1 / 3) + delta_R - delta_B, H)
    H = np.where(B == maximum, (2 / 3) + delta_G - delta_R, H)
    H[np.asarray(H < 0)] += 1
    H[np.asarray(H > 1)] -= 1
    H[np.asarray(delta == 0)] = 0

    HSL = tstack([H, S, L])

    return from_range_1(HSL)


def HSL_to_RGB(HSL: ArrayLike) -> NDArray:
    """
    Convert from *HSL* colourspace to *RGB* colourspace.

    Parameters
    ----------
    HSL
        *HSL* colourspace array.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HSL``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`EasyRGBk`, :cite:`Smith1978b`, :cite:`Wikipedia2003`

    Examples
    --------
    >>> HSL = np.array([0.99603944, 0.87347144, 0.24350795])
    >>> HSL_to_RGB(HSL)  # doctest: +ELLIPSIS
    array([ 0.4562051...,  0.0308107...,  0.0409195...])
    """

    H, S, L = tsplit(to_domain_1(HSL))

    def H_to_RGB(vi: NDArray, vj: NDArray, vH: NDArray) -> NDArray:
        """Convert *hue* value to *RGB* colourspace."""

        vH = as_float_array(vH)

        vH[np.asarray(vH < 0)] += 1
        vH[np.asarray(vH > 1)] -= 1

        v = np.where(
            6 * vH < 1,
            vi + (vj - vi) * 6 * vH,
            np.nan,
        )
        v = np.where(np.logical_and(2 * vH < 1, np.isnan(v)), vj, v)
        v = np.where(
            np.logical_and(3 * vH < 2, np.isnan(v)),
            vi + (vj - vi) * ((2 / 3) - vH) * 6,
            v,
        )
        v = np.where(np.isnan(v), vi, v)

        return v

    j = np.where(L < 0.5, L * (1 + S), (L + S) - (S * L))
    i = 2 * L - j

    R = H_to_RGB(i, j, H + (1 / 3))
    G = H_to_RGB(i, j, H)
    B = H_to_RGB(i, j, H - (1 / 3))

    R = np.where(S == 0, L, R)
    G = np.where(S == 0, L, G)
    B = np.where(S == 0, L, B)

    RGB = tstack([R, G, B])

    return from_range_1(RGB)


def RGB_to_HCL(
    RGB: ArrayLike, gamma: Floating = 3, Y_0: Floating = 100
) -> NDArray:
    """
    Convert from *RGB* colourspace to *HCL* colourspace according to
    *Sarifuddin and Missaoui (2005)* method.

    Parameters
    ----------
    RGB
        *RGB* colourspace array.
    gamma
        Non-linear lightness exponent matching *Lightness* :math:`L^*`.
    Y_0
        White reference luminance :math:`Y_0`.

    Returns
    -------
    :class:`numpy.ndarray`
        *HCL* array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HCL``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Sarifuddin2005`, :cite:`Wikipedia2015`

    Examples
    --------
    >>> RGB = np.array([0.45620519, 0.03081071, 0.04091952])
    >>> RGB_to_HCL(RGB)  # doctest: +ELLIPSIS
    array([-0.0316785...,  0.2841715...,  0.2285964...])
    """

    R, G, B = tsplit(to_domain_1(RGB))

    Min = np.minimum(np.minimum(R, G), B)
    Max = np.maximum(np.maximum(R, G), B)

    alpha = (Min / Max) / Y_0

    Q = np.exp(alpha * gamma)

    L = (Q * Max + (Q - 1) * Min) / 2

    R_G = R - G
    G_B = G - B
    B_R = B - R

    C = Q * (np.abs(R_G) + np.abs(G_B) + np.abs(B_R)) / 3

    H = np.arctan(G_B / R_G)

    _2_3_H = 2 / 3 * H
    _4_3_H = 4 / 3 * H

    H = np.select(
        [
            np.logical_and(R_G >= 0, G_B >= 0),
            np.logical_and(R_G >= 0, G_B < 0),
            np.logical_and(R_G < 0, G_B >= 0),
            np.logical_and(R_G < 0, G_B < 0),
        ],
        [
            _2_3_H,
            _4_3_H,
            np.pi + _4_3_H,
            _2_3_H - np.pi,
        ],
    )

    HCL = tstack([H, C, L])

    return from_range_1(HCL)


def HCL_to_RGB(
    HCL: ArrayLike, gamma: Floating = 3, Y_0: Floating = 100
) -> NDArray:
    """
    Convert from *HCL* colourspace to *RGB* colourspace according to
    *Sarifuddin and Missaoui (2005)* method.

    Parameters
    ----------
    HCL
        *HCL* colourspace array.
    gamma
        Non-linear lightness exponent matching *Lightness* :math:`L^*`.
    Y_0
        White reference luminance :math:`Y_0`.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace array.

    Notes
    -----
    +------------+-----------------------+---------------+
    | **Domain** | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``HCL``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    +------------+-----------------------+---------------+
    | **Range**  | **Scale - Reference** | **Scale - 1** |
    +============+=======================+===============+
    | ``RGB``    | [0, 1]                | [0, 1]        |
    +------------+-----------------------+---------------+

    References
    ----------
    :cite:`Sarifuddin2005`, :cite:`Wikipedia2015`

    Examples
    --------
    >>> HCL = np.array([-0.03167854, 0.28417150, 0.22859647])
    >>> HCL_to_RGB(HCL)  # doctest: +ELLIPSIS
    array([ 0.4562033...,  0.0308104...,  0.0409192...])
    """

    H, C, L = tsplit(to_domain_1(HCL))

    Q = np.exp((1 - (3 * C) / (4 * L)) * (gamma / Y_0))

    Min = (4 * L - 3 * C) / (4 * Q - 2)
    Max = Min + (3 * C) / (2 * Q)

    def _1_2_3(a: ArrayLike) -> NDArray:
        """Tail-stack given :math:`a` array as a *bool* dtype."""

        return tstack([a, a, a], dtype=np.bool_)

    tan_3_2_H = np.tan(3 / 2 * H)
    tan_3_4_H_MP = np.tan(3 / 4 * (H - np.pi))
    tan_3_4_H = np.tan(3 / 4 * H)
    tan_3_2_H_PP = np.tan(3 / 2 * (H + np.pi))

    RGB = np.select(
        [
            _1_2_3(np.logical_and(0 <= H, H <= np.radians(60))),
            _1_2_3(np.logical_and(np.radians(60) < H, H <= np.radians(120))),
            _1_2_3(np.logical_and(np.radians(120) < H, H <= np.pi)),
            _1_2_3(np.logical_and(np.radians(-60) <= H, H < 0)),
            _1_2_3(np.logical_and(np.radians(-120) <= H, H < np.radians(-60))),
            _1_2_3(np.logical_and(-np.pi < H, H < np.radians(-120))),
        ],
        [
            tstack(
                [
                    Max,
                    (Max * tan_3_2_H + Min) / (1 + tan_3_2_H),
                    Min,
                ]
            ),
            tstack(
                [
                    (Max * (1 + tan_3_4_H_MP) - Min) / tan_3_4_H_MP,
                    Max,
                    Min,
                ]
            ),
            tstack(
                [
                    Min,
                    Max,
                    Max * (1 + tan_3_4_H_MP) - Min * tan_3_4_H_MP,
                ]
            ),
            tstack(
                [
                    Max,
                    Min,
                    Min * (1 + tan_3_4_H) - Max * tan_3_4_H,
                ]
            ),
            tstack(
                [
                    (Min * (1 + tan_3_4_H) - Max) / tan_3_4_H,
                    Min,
                    Max,
                ]
            ),
            tstack(
                [
                    Min,
                    (Min * tan_3_2_H_PP + Max) / (1 + tan_3_2_H_PP),
                    Max,
                ]
            ),
        ],
    )

    return from_range_1(RGB)



"""Showcases *ICtCp* *colour encoding* computations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box('"ICtCp" Colour Encoding Computations')

RGB = np.array([0.45620519, 0.03081071, 0.04091952])
message_box(
    f'Converting from the "ITU-R BT.2020" colourspace to the "ICtCp" colour '
    f'encoding given "RGB" values:\n\n\t{RGB}'
)
print(colour.RGB_to_ICtCp(RGB))

print("\n")

ICtCp = np.array([0.07351364, 0.00475253, 0.09351596])
message_box(
    f'Converting from the "ICtCp" colour encoding to the "ITU-R BT.2020" '
    f'colourspace given "ICtCp" values:\n\n\t{ICtCp}'
)
print(colour.ICtCp_to_RGB(ICtCp))



from .rayleigh import (
    scattering_cross_section,
    rayleigh_optical_depth,
    rayleigh_scattering,
    sd_rayleigh_scattering,
)

__all__ = [
    "scattering_cross_section",
    "rayleigh_optical_depth",
    "rayleigh_scattering",
    "sd_rayleigh_scattering",
]



from colour.utilities import is_matplotlib_installed

is_matplotlib_installed(raise_exception=True)

from .datasets import *  # noqa
from . import datasets  # noqa
from .common import (  # noqa
    CONSTANTS_COLOUR_STYLE,
    CONSTANTS_ARROW_STYLE,
    colour_style,
    override_style,
    XYZ_to_plotting_colourspace,
    ColourSwatch,
    colour_cycle,
    artist,
    camera,
    render,
    label_rectangles,
    uniform_axes3d,
    filter_passthrough,
    filter_RGB_colourspaces,
    filter_cmfs,
    filter_illuminants,
    filter_colour_checkers,
    update_settings_collection,
    plot_single_colour_swatch,
    plot_multi_colour_swatches,
    plot_single_function,
    plot_multi_functions,
    plot_image,
)
from .blindness import plot_cvd_simulation_Machado2009  # noqa
from .colorimetry import (  # noqa
    plot_single_sd,
    plot_multi_sds,
    plot_single_cmfs,
    plot_multi_cmfs,
    plot_single_illuminant_sd,
    plot_multi_illuminant_sds,
    plot_visible_spectrum,
    plot_single_lightness_function,
    plot_multi_lightness_functions,
    plot_single_luminance_function,
    plot_multi_luminance_functions,
    plot_blackbody_spectral_radiance,
    plot_blackbody_colours,
)
from .characterisation import (  # noqa
    plot_single_colour_checker,
    plot_multi_colour_checkers,
)
from .diagrams import (  # noqa
    plot_chromaticity_diagram_CIE1931,
    plot_chromaticity_diagram_CIE1960UCS,
    plot_chromaticity_diagram_CIE1976UCS,
    plot_sds_in_chromaticity_diagram_CIE1931,
    plot_sds_in_chromaticity_diagram_CIE1960UCS,
    plot_sds_in_chromaticity_diagram_CIE1976UCS,
)
from .corresponding import plot_corresponding_chromaticities_prediction  # noqa
from .graph import plot_automatic_colour_conversion_graph  # noqa
from .models import (  # noqa
    colourspace_model_axis_reorder,
    plot_pointer_gamut,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS,
    plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS,
    plot_single_cctf,
    plot_multi_cctfs,
    plot_constant_hue_loci,
)
from .notation import (  # noqa
    plot_single_munsell_value_function,
    plot_multi_munsell_value_functions,
)
from .phenomena import (  # noqa
    plot_single_sd_rayleigh_scattering,
    plot_the_blue_sky,
)
from .quality import (  # noqa
    plot_single_sd_colour_rendering_index_bars,
    plot_multi_sds_colour_rendering_indexes_bars,
    plot_single_sd_colour_quality_scale_bars,
    plot_multi_sds_colour_quality_scales_bars,
)
from .section import (  # noqa
    plot_visible_spectrum_section,
    plot_RGB_colourspace_section,
)
from .temperature import (  # noqa
    plot_planckian_locus_in_chromaticity_diagram_CIE1931,
    plot_planckian_locus_in_chromaticity_diagram_CIE1960UCS,
)
from .tm3018 import plot_single_sd_colour_rendition_report  # noqa
from .volume import plot_RGB_colourspaces_gamuts, plot_RGB_scatter  # noqa

__all__ = []
__all__ += datasets.__all__
__all__ += [
    "CONSTANTS_COLOUR_STYLE",
    "CONSTANTS_ARROW_STYLE",
    "colour_style",
    "override_style",
    "XYZ_to_plotting_colourspace",
    "ColourSwatch",
    "colour_cycle",
    "artist",
    "camera",
    "render",
    "label_rectangles",
    "uniform_axes3d",
    "filter_passthrough",
    "filter_RGB_colourspaces",
    "filter_cmfs",
    "filter_illuminants",
    "filter_colour_checkers",
    "update_settings_collection",
    "plot_single_colour_swatch",
    "plot_multi_colour_swatches",
    "plot_single_function",
    "plot_multi_functions",
    "plot_image",
]
__all__ += [
    "plot_cvd_simulation_Machado2009",
]
__all__ += [
    "plot_single_sd",
    "plot_multi_sds",
    "plot_single_cmfs",
    "plot_multi_cmfs",
    "plot_single_illuminant_sd",
    "plot_multi_illuminant_sds",
    "plot_visible_spectrum",
    "plot_single_lightness_function",
    "plot_multi_lightness_functions",
    "plot_single_luminance_function",
    "plot_multi_luminance_functions",
    "plot_blackbody_spectral_radiance",
    "plot_blackbody_colours",
]
__all__ += [
    "plot_single_colour_checker",
    "plot_multi_colour_checkers",
]
__all__ += [
    "plot_chromaticity_diagram_CIE1931",
    "plot_chromaticity_diagram_CIE1960UCS",
    "plot_chromaticity_diagram_CIE1976UCS",
    "plot_sds_in_chromaticity_diagram_CIE1931",
    "plot_sds_in_chromaticity_diagram_CIE1960UCS",
    "plot_sds_in_chromaticity_diagram_CIE1976UCS",
]
__all__ += [
    "plot_corresponding_chromaticities_prediction",
]
__all__ += [
    "plot_automatic_colour_conversion_graph",
]
__all__ += [
    "colourspace_model_axis_reorder",
    "plot_pointer_gamut",
    "plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931",
    "plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS",
    "plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS",
    "plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931",
    "plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS",
    "plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS",
    "plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1931",
    "plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1960UCS",
    "plot_ellipses_MacAdam1942_in_chromaticity_diagram_CIE1976UCS",
    "plot_single_cctf",
    "plot_multi_cctfs",
    "plot_constant_hue_loci",
]
__all__ += [
    "plot_single_munsell_value_function",
    "plot_multi_munsell_value_functions",
]
__all__ += [
    "plot_single_sd_rayleigh_scattering",
    "plot_the_blue_sky",
]
__all__ += [
    "plot_single_sd_colour_rendering_index_bars",
    "plot_multi_sds_colour_rendering_indexes_bars",
    "plot_single_sd_colour_quality_scale_bars",
    "plot_multi_sds_colour_quality_scales_bars",
]
__all__ += [
    "plot_visible_spectrum_section",
    "plot_RGB_colourspace_section",
]
__all__ += [
    "plot_planckian_locus_in_chromaticity_diagram_CIE1931",
    "plot_planckian_locus_in_chromaticity_diagram_CIE1960UCS",
]
__all__ += [
    "plot_single_sd_colour_rendition_report",
]
__all__ += [
    "plot_RGB_colourspaces_gamuts",
    "plot_RGB_scatter",
]



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.appearance.hunt` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.appearance import (
    VIEWING_CONDITIONS_HUNT,
    InductionFactors_Hunt,
    XYZ_to_Hunt,
)
from colour.utilities import (
    as_float_array,
    domain_range_scale,
    ignore_numpy_errors,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_Hunt",
]


class TestXYZ_to_Hunt(unittest.TestCase):
    """
    Define :func:`colour.appearance.hunt.XYZ_to_Hunt` definition unit tests
    methods.
    """

    def test_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition.

        Notes
        -----
        -   The test values have been generated from data of the following file
            by *Fairchild (2013)*:
            http://rit-mcsl.org/fairchild//files/AppModEx.xls
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        XYZ_b = XYZ_w * np.array([1, 0.2, 1])
        L_A = 318.31
        surround = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
        CCT_w = 6504.0
        np.testing.assert_allclose(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            np.array([42.12, 0.16, 269.3, 0.03, 31.92, 0.16, np.nan, np.nan]),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([57.06, 43.06, 31.96])
        L_A = 31.83
        np.testing.assert_allclose(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            np.array(
                [66.76, 63.89, 18.6, 153.36, 31.22, 58.28, np.nan, np.nan]
            ),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([3.53, 6.56, 2.14])
        XYZ_w = np.array([109.85, 100.00, 35.58])
        L_A = 318.31
        CCT_w = 2856
        np.testing.assert_allclose(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            np.array(
                [19.56, 74.58, 178.3, 245.4, 18.9, 76.33, np.nan, np.nan]
            ),
            rtol=0.01,
            atol=0.01,
        )

        XYZ = np.array([19.01, 20.00, 21.78])
        L_A = 31.83
        np.testing.assert_allclose(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            np.array(
                [40.27, 73.84, 262.8, 209.29, 22.15, 67.35, np.nan, np.nan]
            ),
            rtol=0.01,
            atol=0.01,
        )

    def test_n_dimensional_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition
        n-dimensional support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        XYZ_b = XYZ_w * np.array([1, 0.2, 1])
        L_A = 318.31
        surround = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
        CCT_w = 6504.0
        specification = XYZ_to_Hunt(
            XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w
        )

        XYZ = np.tile(XYZ, (6, 1))
        specification = np.tile(specification, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            specification,
            decimal=7,
        )

        XYZ_w = np.tile(XYZ_w, (6, 1))
        XYZ_b = np.tile(XYZ_b, (6, 1))
        np.testing.assert_almost_equal(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            specification,
            decimal=7,
        )

        XYZ = np.reshape(XYZ, (2, 3, 3))
        XYZ_w = np.reshape(XYZ_w, (2, 3, 3))
        XYZ_b = np.reshape(XYZ_b, (2, 3, 3))
        specification = np.reshape(specification, (2, 3, 8))
        np.testing.assert_almost_equal(
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w),
            specification,
            decimal=7,
        )

    def test_domain_range_scale_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition domain
        and range scale support.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        XYZ_b = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        surround = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
        CCT_w = 6504.0
        specification = XYZ_to_Hunt(
            XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w
        )

        d_r = (
            ("reference", 1, 1),
            ("1", 0.01, np.array([1, 1, 1 / 360, 1, 1, 1, np.nan, np.nan])),
            ("100", 1, np.array([1, 1, 100 / 360, 1, 1, 1, np.nan, np.nan])),
        )
        for scale, factor_a, factor_b in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_Hunt(
                        XYZ * factor_a,
                        XYZ_w * factor_a,
                        XYZ_b * factor_a,
                        L_A,
                        surround,
                        CCT_w=CCT_w,
                    ),
                    as_float_array(specification) * factor_b,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_raise_exception_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition raised
        exception.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        XYZ_b = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        surround = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
        CCT_w = 6504.0
        S = S_w = 0.5

        try:
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround)
        except ValueError:
            pass

        try:
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w, S=S)
        except ValueError:
            pass

        try:
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w, S_w=S_w)
        except ValueError:
            pass

    @ignore_numpy_errors
    def test_XYZ_p_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition *XYZ_p*
        argument handling.
        """

        XYZ = np.array([19.01, 20.00, 21.78])
        XYZ_w = np.array([95.05, 100.00, 108.88])
        XYZ_b = XYZ_p = np.array([95.05, 100.00, 108.88])
        L_A = 318.31
        surround = VIEWING_CONDITIONS_HUNT["Normal Scenes"]
        CCT_w = 6504.0

        np.testing.assert_almost_equal(
            XYZ_to_Hunt(
                XYZ,
                XYZ_w,
                XYZ_b,
                L_A,
                surround,
                XYZ_p=XYZ_p,
                CCT_w=CCT_w,
            ),
            np.array(
                [
                    30.046267861960700,
                    0.121050839936350,
                    269.273759446144600,
                    0.019909320692942,
                    22.209765491265024,
                    0.123896438259997,
                    np.nan,
                    np.nan,
                ]
            ),
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_XYZ_to_Hunt(self):
        """
        Test :func:`colour.appearance.hunt.XYZ_to_Hunt` definition
        nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_w = np.array(case)
            XYZ_b = np.array(case)
            L_A = case[0]
            surround = InductionFactors_Hunt(case[0], case[0])
            CCT_w = case[0]
            XYZ_to_Hunt(XYZ, XYZ_w, XYZ_b, L_A, surround, CCT_w=CCT_w)



from .abstract import AbstractContinuousFunction
from .signal import Signal
from .multi_signals import MultiSignals

__all__ = []
__all__ += [
    "AbstractContinuousFunction",
]
__all__ += [
    "Signal",
]
__all__ += [
    "MultiSignals",
]



# !/usr/bin/env python
"""Defines the unit tests for the :mod:`colour.adaptation.fairchild1990` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.adaptation import chromatic_adaptation_Fairchild1990
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestChromaticAdaptationFairchild1990",
]


class TestChromaticAdaptationFairchild1990(unittest.TestCase):
    """
    Define :func:`colour.adaptation.fairchild1990.\
chromatic_adaptation_Fairchild1990` definition unit tests methods.
    """

    def test_chromatic_adaptation_Fairchild1990(self):
        """
        Test :func:`colour.adaptation.fairchild1990.\
chromatic_adaptation_Fairchild1990` definition.
        """

        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(
                np.array([19.53, 23.07, 24.97]),
                np.array([111.15, 100.00, 35.20]),
                np.array([94.81, 100.00, 107.30]),
                200,
            ),
            np.array([23.32526349, 23.32455819, 76.11593750]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(
                np.array([0.14222010, 0.23042768, 0.10495772]) * 100,
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([1.09846607, 1.00000000, 0.35582280]) * 100,
                200,
            ),
            np.array([19.28089326, 22.91583715, 3.42923503]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(
                np.array([0.07818780, 0.06157201, 0.28099326]) * 100,
                np.array([0.95045593, 1.00000000, 1.08905775]) * 100,
                np.array([0.99144661, 1.00000000, 0.67315942]) * 100,
                200,
            ),
            np.array([6.35093475, 6.13061347, 17.36852430]),
            decimal=7,
        )

    def test_n_dimensional_chromatic_adaptation_Fairchild1990(self):
        """
        Test :func:`colour.adaptation.fairchild1990.\
chromatic_adaptation_Fairchild1990` definition n-dimensional arrays support.
        """

        XYZ_1 = np.array([19.53, 23.07, 24.97])
        XYZ_n = np.array([111.15, 100.00, 35.20])
        XYZ_r = np.array([94.81, 100.00, 107.30])
        Y_n = 200
        XYZ_c = chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n)

        XYZ_1 = np.tile(XYZ_1, (6, 1))
        XYZ_c = np.tile(XYZ_c, (6, 1))
        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n),
            XYZ_c,
            decimal=7,
        )

        XYZ_n = np.tile(XYZ_n, (6, 1))
        XYZ_r = np.tile(XYZ_r, (6, 1))
        Y_n = np.tile(Y_n, 6)
        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n),
            XYZ_c,
            decimal=7,
        )

        XYZ_1 = np.reshape(XYZ_1, (2, 3, 3))
        XYZ_n = np.reshape(XYZ_n, (2, 3, 3))
        XYZ_r = np.reshape(XYZ_r, (2, 3, 3))
        Y_n = np.reshape(Y_n, (2, 3))
        XYZ_c = np.reshape(XYZ_c, (2, 3, 3))
        np.testing.assert_almost_equal(
            chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n),
            XYZ_c,
            decimal=7,
        )

    def test_domain_range_scale_chromatic_adaptation_Fairchild1990(self):
        """
        Test :func:`colour.adaptation.fairchild1990.\
chromatic_adaptation_Fairchild1990` definition domain and range scale support.
        """

        XYZ_1 = np.array([19.53, 23.07, 24.97])
        XYZ_n = np.array([111.15, 100.00, 35.20])
        XYZ_r = np.array([94.81, 100.00, 107.30])
        Y_n = 200
        XYZ_c = chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n)

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    chromatic_adaptation_Fairchild1990(
                        XYZ_1 * factor, XYZ_n * factor, XYZ_r * factor, Y_n
                    ),
                    XYZ_c * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_chromatic_adaptation_Fairchild1990(self):
        """
        Test :func:`colour.adaptation.fairchild1990.\
chromatic_adaptation_Fairchild1990` definition nan support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ_1 = np.array(case)
            XYZ_n = np.array(case)
            XYZ_r = np.array(case)
            Y_n = case[0]
            chromatic_adaptation_Fairchild1990(XYZ_1, XYZ_n, XYZ_r, Y_n)


if __name__ == "__main__":
    unittest.main()



"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.canon_log` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    log_encoding_CanonLog,
    log_decoding_CanonLog,
    log_encoding_CanonLog2,
    log_decoding_CanonLog2,
    log_encoding_CanonLog3,
    log_decoding_CanonLog3,
)

from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLogEncoding_CanonLog",
    "TestLogDecoding_CanonLog",
    "TestLogEncoding_CanonLog2",
    "TestLogDecoding_CanonLog2",
    "TestLogEncoding_CanonLog3",
    "TestLogDecoding_CanonLog3",
]


class TestLogEncoding_CanonLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog` definition unit tests methods.
    """

    def test_log_encoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog` definition.
        """

        self.assertAlmostEqual(
            log_encoding_CanonLog(-0.1), -0.023560122781997, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(0.0), 0.125122480156403, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(0.18), 0.343389651726069, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(0.18, 12), 0.343138084215647, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(0.18, 10, False), 0.327953896935809, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(0.18, 10, False, False),
            0.312012855550395,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog(1.0), 0.618775485598649, places=7
        )

    def test_n_dimensional_log_encoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog` definition n-dimensional arrays support.
        """

        x = 0.18
        clog = log_encoding_CanonLog(x)

        x = np.tile(x, 6)
        clog = np.tile(clog, 6)
        np.testing.assert_almost_equal(
            log_encoding_CanonLog(x), clog, decimal=7
        )

        x = np.reshape(x, (2, 3))
        clog = np.reshape(clog, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog(x), clog, decimal=7
        )

        x = np.reshape(x, (2, 3, 1))
        clog = np.reshape(clog, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog(x), clog, decimal=7
        )

    def test_domain_range_scale_log_encoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog` definition domain and range scale support.
        """

        x = 0.18
        clog = log_encoding_CanonLog(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_CanonLog(x * factor), clog * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog` definition nan support.
        """

        log_encoding_CanonLog(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_CanonLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog` definition unit tests methods.
    """

    def test_log_decoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog` definition.
        """

        self.assertAlmostEqual(
            log_decoding_CanonLog(-0.023560122781997), -0.1, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.125122480156403), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.343389651726069), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.343138084215647, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.327953896935809, 10, False), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.312012855550395, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog(0.618775485598649), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog` definition n-dimensional arrays support.
        """

        clog = 0.343389651726069
        x = log_decoding_CanonLog(clog)

        clog = np.tile(clog, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(
            log_decoding_CanonLog(clog), x, decimal=7
        )

        clog = np.reshape(clog, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog(clog), x, decimal=7
        )

        clog = np.reshape(clog, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog(clog), x, decimal=7
        )

    def test_domain_range_scale_log_decoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog` definition domain and range scale support.
        """

        clog = 0.343389651726069
        x = log_decoding_CanonLog(clog)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_CanonLog(clog * factor), x * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_CanonLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog` definition nan support.
        """

        log_decoding_CanonLog(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogEncoding_CanonLog2(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog2` definition unit tests methods.
    """

    def test_log_encoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog2` definition.
        """

        self.assertAlmostEqual(
            log_encoding_CanonLog2(-0.1), -0.155370131996824, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(0.0), 0.092864125247312, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(0.18), 0.398254694983167, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(0.18, 12), 0.397962933301861, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(0.18, 10, False),
            0.392025745397009,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(0.18, 10, False, False),
            0.379864582222983,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog2(1.0), 0.573229282897641, places=7
        )

    def test_n_dimensional_log_encoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog2` definition n-dimensional arrays support.
        """

        x = 0.18
        clog2 = log_encoding_CanonLog2(x)

        x = np.tile(x, 6)
        clog2 = np.tile(clog2, 6)
        np.testing.assert_almost_equal(
            log_encoding_CanonLog2(x), clog2, decimal=7
        )

        x = np.reshape(x, (2, 3))
        clog2 = np.reshape(clog2, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog2(x), clog2, decimal=7
        )

        x = np.reshape(x, (2, 3, 1))
        clog2 = np.reshape(clog2, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog2(x), clog2, decimal=7
        )

    def test_domain_range_scale_log_encoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog2` definition domain and range scale support.
        """

        x = 0.18
        clog2 = log_encoding_CanonLog2(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_CanonLog2(x * factor),
                    clog2 * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog2` definition nan support.
        """

        log_encoding_CanonLog2(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_CanonLog2(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog2` definition unit tests methods.
    """

    def test_log_decoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog2` definition.
        """

        self.assertAlmostEqual(
            log_decoding_CanonLog2(-0.155370131996824), -0.1, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.092864125247312), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.398254694983167), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.397962933301861, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.392025745397009, 10, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.379864582222983, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog2(0.573229282897641), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog2` definition n-dimensional arrays support.
        """

        clog2 = 0.398254694983167
        x = log_decoding_CanonLog2(clog2)

        clog2 = np.tile(clog2, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(
            log_decoding_CanonLog2(clog2), x, decimal=7
        )

        clog2 = np.reshape(clog2, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog2(clog2), x, decimal=7
        )

        clog2 = np.reshape(clog2, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog2(clog2), x, decimal=7
        )

    def test_domain_range_scale_log_decoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog2` definition domain and range scale support.
        """

        clog = 0.398254694983167
        x = log_decoding_CanonLog2(clog)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_CanonLog2(clog * factor),
                    x * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_CanonLog2(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog2` definition nan support.
        """

        log_decoding_CanonLog2(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogEncoding_CanonLog3(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog3` definition unit tests methods.
    """

    def test_log_encoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog3` definition.
        """

        self.assertAlmostEqual(
            log_encoding_CanonLog3(-0.1), -0.028494506076432, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(0.0), 0.125122189869013, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(0.18), 0.343389369388687, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(0.18, 12), 0.343137802085105, places=7
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(0.18, 10, False),
            0.327953567219893,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(0.18, 10, False, False),
            0.313436005886328,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_CanonLog3(1.0), 0.580277796238604, places=7
        )

    def test_n_dimensional_log_encoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog3` definition n-dimensional arrays support.
        """

        x = 0.18
        clog3 = log_encoding_CanonLog3(x)

        x = np.tile(x, 6)
        clog3 = np.tile(clog3, 6)
        np.testing.assert_almost_equal(
            log_encoding_CanonLog3(x), clog3, decimal=7
        )

        x = np.reshape(x, (2, 3))
        clog3 = np.reshape(clog3, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog3(x), clog3, decimal=7
        )

        x = np.reshape(x, (2, 3, 1))
        clog3 = np.reshape(clog3, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_CanonLog3(x), clog3, decimal=7
        )

    def test_domain_range_scale_log_encoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog3` definition domain and range scale support.
        """

        x = 0.18
        clog3 = log_encoding_CanonLog3(x)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_CanonLog3(x * factor),
                    clog3 * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_encoding_CanonLog3` definition nan support.
        """

        log_encoding_CanonLog3(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


class TestLogDecoding_CanonLog3(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog3` definition unit tests methods.
    """

    def test_log_decoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog3` definition.
        """

        self.assertAlmostEqual(
            log_decoding_CanonLog3(-0.028494506076432), -0.1, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.125122189869013), 0.0, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.343389369388687), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.343137802085105, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.327953567219893, 10, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.313436005886328, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_CanonLog3(0.580277796238604), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog3` definition n-dimensional arrays support.
        """

        clog3 = 0.343389369388687
        x = log_decoding_CanonLog3(clog3)

        clog3 = np.tile(clog3, 6)
        x = np.tile(x, 6)
        np.testing.assert_almost_equal(
            log_decoding_CanonLog3(clog3), x, decimal=7
        )

        clog3 = np.reshape(clog3, (2, 3))
        x = np.reshape(x, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog3(clog3), x, decimal=7
        )

        clog3 = np.reshape(clog3, (2, 3, 1))
        x = np.reshape(x, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_CanonLog3(clog3), x, decimal=7
        )

    def test_domain_range_scale_log_decoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog3` definition domain and range scale support.
        """

        clog = 0.343389369388687
        x = log_decoding_CanonLog3(clog)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_CanonLog3(clog * factor),
                    x * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_CanonLog3(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.canon_log.\
log_decoding_CanonLog3` definition nan support.
        """

        log_decoding_CanonLog3(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


if __name__ == "__main__":
    unittest.main()



"""Showcases Cyan-Magenta-Yellow (Black) (CMY(K)) colour transformations."""

import numpy as np

import colour
from colour.utilities import message_box

message_box("Cyan-Magenta-Yellow (Black) (CMY(K)) Colour Transformations")

RGB = np.array([0.45620519, 0.03081071, 0.04091952])
message_box(
    f'Converting to the "CMY" colourspace from given "RGB" colourspace '
    f"values:\n\n\t{RGB}"
)
print(colour.RGB_to_CMY(RGB))

print("\n")

CMY = np.array([0.54379481, 0.96918929, 0.95908048])
message_box(
    f'Converting to the "RGB" colourspace from given "CMY" colourspace '
    f"values:\n\n\t{CMY}"
)
print(colour.CMY_to_RGB(CMY))

print("\n")

message_box(
    f'Converting to the "CMYK" colourspace from given "CMY" colourspace '
    f"values:\n\n\t{CMY}"
)
print(colour.CMY_to_CMYK(CMY))

print("\n")

CMYK = np.array([0.00000000, 0.93246304, 0.91030457, 0.54379481])
message_box(
    f'Converting to the "CMY" colourspace from given "CMYK" colourspace '
    f"values:\n\n\t{CMYK}"
)
print(colour.CMYK_to_CMY(CMYK))



"""
Optical Phenomenon Plotting
===========================

Defines the optical phenomena plotting objects:

-   :func:`colour.plotting.plot_single_sd_rayleigh_scattering`
-   :func:`colour.plotting.plot_the_blue_sky`
"""

from __future__ import annotations

import matplotlib.pyplot as plt

from colour.algebra import normalise_maximum
from colour.colorimetry import (
    MultiSpectralDistributions,
    SpectralDistribution,
    sd_to_XYZ,
)
from colour.hints import (
    Any,
    Dict,
    FloatingOrArrayLike,
    Sequence,
    Tuple,
    Union,
    cast,
)
from colour.phenomena import sd_rayleigh_scattering
from colour.phenomena.rayleigh import (
    CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL,
    CONSTANT_DEFAULT_ALTITUDE,
    CONSTANT_DEFAULT_LATITUDE,
    CONSTANT_STANDARD_AIR_TEMPERATURE,
    CONSTANT_STANDARD_CO2_CONCENTRATION,
)
from colour.plotting import (
    SD_ASTMG173_ETR,
    CONSTANTS_COLOUR_STYLE,
    ColourSwatch,
    XYZ_to_plotting_colourspace,
    filter_cmfs,
    override_style,
    render,
    plot_single_colour_swatch,
    plot_single_sd,
)
from colour.utilities import first_item

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "plot_single_sd_rayleigh_scattering",
    "plot_the_blue_sky",
]


@override_style()
def plot_single_sd_rayleigh_scattering(
    CO2_concentration: FloatingOrArrayLike = CONSTANT_STANDARD_CO2_CONCENTRATION,
    temperature: FloatingOrArrayLike = CONSTANT_STANDARD_AIR_TEMPERATURE,
    pressure: FloatingOrArrayLike = CONSTANT_AVERAGE_PRESSURE_MEAN_SEA_LEVEL,
    latitude: FloatingOrArrayLike = CONSTANT_DEFAULT_LATITUDE,
    altitude: FloatingOrArrayLike = CONSTANT_DEFAULT_ALTITUDE,
    cmfs: Union[
        MultiSpectralDistributions,
        str,
        Sequence[Union[MultiSpectralDistributions, str]],
    ] = "CIE 1931 2 Degree Standard Observer",
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot a single *Rayleigh* scattering spectral distribution.

    Parameters
    ----------
    CO2_concentration
        :math:`CO_2` concentration in parts per million (ppm).
    temperature
        Air temperature :math:`T[K]` in kelvin degrees.
    pressure
        Surface pressure :math:`P` of the measurement site.
    latitude
        Latitude of the site in degrees.
    altitude
        Altitude of the site in meters.
    cmfs
        Standard observer colour matching functions used for computing the
        spectrum domain and colours. ``cmfs`` can be of any type or form
        supported by the :func:`colour.plotting.filter_cmfs` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_single_sd`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_single_sd_rayleigh_scattering()  # doctest: +ELLIPSIS
    (<Figure size ... with 1 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_Single_SD_Rayleigh_Scattering.png
        :align: center
        :alt: plot_single_sd_rayleigh_scattering
    """

    title = "Rayleigh Scattering"

    cmfs = cast(
        MultiSpectralDistributions, first_item(filter_cmfs(cmfs).values())
    )

    settings: Dict[str, Any] = {"title": title, "y_label": "Optical Depth"}
    settings.update(kwargs)

    sd = sd_rayleigh_scattering(
        cmfs.shape,
        CO2_concentration,
        temperature,
        pressure,
        latitude,
        altitude,
    )

    return plot_single_sd(sd, **settings)


@override_style()
def plot_the_blue_sky(
    cmfs: Union[
        MultiSpectralDistributions,
        str,
        Sequence[Union[MultiSpectralDistributions, str]],
    ] = "CIE 1931 2 Degree Standard Observer",
    **kwargs: Any,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot the blue sky.

    Parameters
    ----------
    cmfs
        Standard observer colour matching functions used for computing the
        spectrum domain and colours. ``cmfs`` can be of any type or form
        supported by the :func:`colour.plotting.filter_cmfs` definition.

    Other Parameters
    ----------------
    kwargs
        {:func:`colour.plotting.artist`,
        :func:`colour.plotting.plot_single_sd`,
        :func:`colour.plotting.plot_multi_colour_swatches`,
        :func:`colour.plotting.render`},
        See the documentation of the previously listed definitions.

    Returns
    -------
    :class:`tuple`
        Current figure and axes.

    Examples
    --------
    >>> plot_the_blue_sky()  # doctest: +ELLIPSIS
    (<Figure size ... with 2 Axes>, <...AxesSubplot...>)

    .. image:: ../_static/Plotting_Plot_The_Blue_Sky.png
        :align: center
        :alt: plot_the_blue_sky
    """

    figure = plt.figure()

    figure.subplots_adjust(hspace=CONSTANTS_COLOUR_STYLE.geometry.short / 2)

    cmfs = cast(
        MultiSpectralDistributions, first_item(filter_cmfs(cmfs).values())
    )

    ASTMG173_sd = cast(SpectralDistribution, SD_ASTMG173_ETR.copy())
    rayleigh_sd = sd_rayleigh_scattering()
    ASTMG173_sd.align(rayleigh_sd.shape)

    sd = rayleigh_sd * ASTMG173_sd

    axes = figure.add_subplot(211)

    settings: Dict[str, Any] = {
        "axes": axes,
        "title": "The Blue Sky - Synthetic Spectral Distribution",
        "y_label": "W / m-2 / nm-1",
    }
    settings.update(kwargs)
    settings["standalone"] = False

    plot_single_sd(sd, cmfs, **settings)

    axes = figure.add_subplot(212)

    x_label = (
        "The sky is blue because molecules in the atmosphere "
        "scatter shorter wavelengths more than longer ones.\n"
        "The synthetic spectral distribution is computed as "
        "follows: "
        "(ASTM G-173 ETR * Standard Air Rayleigh Scattering)."
    )

    settings = {
        "axes": axes,
        "aspect": None,
        "title": "The Blue Sky - Colour",
        "x_label": x_label,
        "y_label": "",
        "x_ticker": False,
        "y_ticker": False,
    }
    settings.update(kwargs)
    settings["standalone"] = False

    blue_sky_color = XYZ_to_plotting_colourspace(
        sd_to_XYZ(cast(SpectralDistribution, sd))
    )

    figure, axes = plot_single_colour_swatch(
        ColourSwatch(normalise_maximum(blue_sky_color)), **settings
    )

    settings = {"axes": axes, "standalone": True}
    settings.update(kwargs)

    return render(**settings)



"""Defines the unit tests for the :mod:`colour.difference` module."""

import numpy as np
import unittest

from colour.difference import delta_E

from colour.utilities import domain_range_scale

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestDelta_E",
]


class TestDelta_E(unittest.TestCase):
    """Define :func:`colour.difference.delta_E` definition unit tests methods."""

    def test_domain_range_scale_delta_E(self):
        """
        Test :func:`colour.difference.delta_E` definition domain and range
        scale support.
        """

        Lab_1 = np.array([100.00000000, 21.57210357, 272.22819350])
        Lab_2 = np.array([100.00000000, 426.67945353, 72.39590835])

        m = ("CIE 1976", "CIE 1994", "CIE 2000", "CMC", "DIN99")
        v = [delta_E(Lab_1, Lab_2, method) for method in m]

        d_r = (("reference", 1), ("1", 0.01), ("100", 1))
        for method, value in zip(m, v):
            for scale, factor in d_r:
                with domain_range_scale(scale):
                    np.testing.assert_almost_equal(
                        delta_E(Lab_1 * factor, Lab_2 * factor, method),
                        value,
                        decimal=7,
                    )


if __name__ == "__main__":
    unittest.main()



"""
Geometry / Hull Section
=======================

Defines various objects to compute hull sections:

-   :func:`colour.geometry.hull_section`
"""

from __future__ import annotations

import numpy as np

from colour.algebra import linear_conversion
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.hints import (
    ArrayLike,
    Boolean,
    Floating,
    Integer,
    Literal,
    NDArray,
    Union,
)
from colour.utilities import (
    as_float_array,
    as_float_scalar,
    required,
    validate_method,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "edges_to_chord",
    "unique_vertices",
    "close_chord",
    "hull_section",
]


def edges_to_chord(edges: ArrayLike, index: Integer = 0) -> NDArray:
    """
    Convert given edges to a chord, starting at given index.

    Parameters
    ----------
    edges
        Edges to convert to a chord.
    index
        Index to start forming the chord at.

    Returns
    -------
    :class:`numpy.ndarray`
        Chord.

    Examples
    --------
    >>> edges = np.array([
    ...     [[-0.0, -0.5, 0.0], [0.5, -0.5, 0.0]],
    ...     [[-0.5, -0.5, 0.0], [-0.0, -0.5, 0.0]],
    ...     [[0.5, 0.5, 0.0], [-0.0, 0.5, 0.0]],
    ...     [[-0.0, 0.5, 0.0], [-0.5, 0.5, 0.0]],
    ...     [[-0.5, 0.0, -0.0], [-0.5, -0.5, -0.0]],
    ...     [[-0.5, 0.5, -0.0], [-0.5, 0.0, -0.0]],
    ...     [[0.5, -0.5, -0.0], [0.5, 0.0, -0.0]],
    ...     [[0.5, 0.0, -0.0], [0.5, 0.5, -0.0]],
    ... ])
    >>> edges_to_chord(edges)
    array([[-0. , -0.5,  0. ],
           [ 0.5, -0.5,  0. ],
           [ 0.5, -0.5, -0. ],
           [ 0.5,  0. , -0. ],
           [ 0.5,  0. , -0. ],
           [ 0.5,  0.5, -0. ],
           [ 0.5,  0.5,  0. ],
           [-0. ,  0.5,  0. ],
           [-0. ,  0.5,  0. ],
           [-0.5,  0.5,  0. ],
           [-0.5,  0.5, -0. ],
           [-0.5,  0. , -0. ],
           [-0.5,  0. , -0. ],
           [-0.5, -0.5, -0. ],
           [-0.5, -0.5,  0. ],
           [-0. , -0.5,  0. ]])
    """

    edge_list = as_float_array(edges).tolist()

    edges_ordered = [edge_list.pop(index)]
    segment = np.array(edges_ordered[0][1])

    while len(edge_list) > 0:
        edges_array = np.array(edge_list)
        d_0 = np.linalg.norm(edges_array[:, 0, :] - segment, axis=1)
        d_1 = np.linalg.norm(edges_array[:, 1, :] - segment, axis=1)
        d_0_argmin, d_1_argmin = d_0.argmin(), d_1.argmin()

        if d_0[d_0_argmin] < d_1[d_1_argmin]:
            edges_ordered.append(edge_list.pop(d_0_argmin))
            segment = np.array(edges_ordered[-1][1])
        else:
            edges_ordered.append(edge_list.pop(d_1_argmin))
            segment = np.array(edges_ordered[-1][0])

    return as_float_array(edges_ordered).reshape([-1, segment.shape[-1]])


def close_chord(vertices: ArrayLike) -> NDArray:
    """
    Close the chord.

    Parameters
    ----------
    vertices
        Vertices of the chord to close.

    Returns
    -------
    :class:`numpy.ndarray`
        Closed chord.

    Examples
    --------
    >>> close_chord(np.array([[0.0, 0.5, 0.0], [0.0, 0.0, 0.5]]))
    array([[ 0. ,  0.5,  0. ],
           [ 0. ,  0. ,  0.5],
           [ 0. ,  0.5,  0. ]])
    """

    vertices = as_float_array(vertices)

    return np.vstack([vertices, vertices[0]])


def unique_vertices(
    vertices: ArrayLike,
    decimals: Integer = np.finfo(DEFAULT_FLOAT_DTYPE).precision - 1,
) -> NDArray:
    """
    Return the unique vertices from given vertices.

    Parameters
    ----------
    vertices
        Vertices to return the unique vertices from.
    decimals
        Decimals used when rounding the vertices prior to comparison.

    Returns
    -------
    :class:`numpy.ndarray`
        Unique vertices.

    Notes
    -----
    -   The vertices are rounded at given ``decimals``.

    Examples
    --------
    >>> unique_vertices(
    ...     np.array([[0.0, 0.5, 0.0], [0.0, 0.0, 0.5], [0.0, 0.5, 0.0]]))
    array([[ 0. ,  0.5,  0. ],
           [ 0. ,  0. ,  0.5]])
    """

    vertices = as_float_array(vertices)

    unique, indexes = np.unique(
        vertices.round(decimals=decimals), axis=0, return_index=True
    )

    return unique[np.argsort(indexes)]


@required("trimesh")
def hull_section(
    hull: trimesh.Trimesh,  # type: ignore[name-defined]  # noqa
    axis: Union[Literal["+z", "+x", "+y"], str] = "+z",
    origin: Floating = 0.5,
    normalise: Boolean = False,
) -> NDArray:
    """
    Compute the hull section for given axis at given origin.

    Parameters
    ----------
    hull
        *Trimesh* hull.
    axis
        Axis the hull section will be normal to.
    origin
        Coordinate along ``axis`` at which to plot the hull section.
    normalise
        Whether to normalise ``axis`` to the extent of the hull along it.

    Returns
    -------
    :class:`numpy.ndarray`
        Hull section vertices.

    Examples
    --------
    >>> from colour.geometry import primitive_cube
    >>> from colour.utilities import is_trimesh_installed
    >>> vertices, faces, outline = primitive_cube(1, 1, 1, 2, 2, 2)
    >>> if is_trimesh_installed:
    ...     import trimesh
    ...     hull = trimesh.Trimesh(vertices['position'], faces, process=False)
    ...     hull_section(hull, origin=0)
    array([[-0. , -0.5,  0. ],
           [ 0.5, -0.5,  0. ],
           [ 0.5,  0. , -0. ],
           [ 0.5,  0.5, -0. ],
           [-0. ,  0.5,  0. ],
           [-0.5,  0.5,  0. ],
           [-0.5,  0. , -0. ],
           [-0.5, -0.5, -0. ],
           [-0. , -0.5,  0. ]])
    """

    import trimesh

    axis = validate_method(
        axis,
        ["+z", "+x", "+y"],
        '"{0}" axis is invalid, it must be one of {1}!',
    )

    if axis == "+x":
        normal, plane = np.array([1, 0, 0]), np.array([origin, 0, 0])
    elif axis == "+y":
        normal, plane = np.array([0, 1, 0]), np.array([0, origin, 0])
    elif axis == "+z":
        normal, plane = np.array([0, 0, 1]), np.array([0, 0, origin])

    if normalise:
        vertices = hull.vertices * normal
        origin = as_float_scalar(
            linear_conversion(
                origin, [0, 1], [np.min(vertices), np.max(vertices)]
            )
        )
        plane[plane != 0] = origin

    section = trimesh.intersections.mesh_plane(hull, normal, plane)
    if len(section) == 0:
        raise ValueError(
            f'No section exists on "{axis}" axis at {origin} origin!'
        )
    section = close_chord(unique_vertices(edges_to_chord(section)))

    return section



"""Showcases *IES TM-27-14* spectral data *XML* files input / output examples."""

import os

import colour
from colour.utilities import message_box

RESOURCES_DIRECTORY = os.path.join(os.path.dirname(__file__), "resources")

message_box('"IES TM-27-14" Spectral Data "XML" File IO')

message_box('Reading spectral data from "IES TM-27-14" "XML" file.')
sd_tm2714 = colour.SpectralDistribution_IESTM2714(
    os.path.join(RESOURCES_DIRECTORY, "TM27 Sample Spectral Data.spdx")
).read()
print(sd_tm2714)

print("\n")

message_box('"IES TM-27-14" spectral data "XML" file header:')
print(f"Manufacturer: {sd_tm2714.header.manufacturer}")
print(f"Catalog Number: {sd_tm2714.header.catalog_number}")
print(f"Description: {sd_tm2714.header.description}")
print(f"Document Creator: {sd_tm2714.header.document_creator}")
print(f"Unique Identifier: {sd_tm2714.header.unique_identifier}")
print(f"Measurement Equipment: {sd_tm2714.header.measurement_equipment}")
print(f"Laboratory: {sd_tm2714.header.laboratory}")
print(f"Report Number: {sd_tm2714.header.report_number}")
print(f"Report Date: {sd_tm2714.header.report_date}")
print(f"Document Creation Date: {sd_tm2714.header.document_creation_date}")
print(f"Comments: {sd_tm2714.header.comments}")

print("\n")

message_box('"IES TM-27-14" spectral data "XML" file spectral distribution:')
print(f"Spectral Quantity: {sd_tm2714.spectral_quantity}")
print(f"Reflection Geometry: {sd_tm2714.reflection_geometry}")
print(f"Transmission Geometry: {sd_tm2714.transmission_geometry}")
print(f"Bandwidth FWHM: {sd_tm2714.bandwidth_FWHM}")
print(f"Bandwidth Corrected: {sd_tm2714.bandwidth_corrected}")

print("\n")

message_box('"IES TM-27-14" spectral data "XML" file spectral data:')
print(sd_tm2714)



"""
RGB Colourspace Volume Computation
==================================

Defines various RGB colourspace volume computation objects:

-   :func:`colour.RGB_colourspace_limits`
-   :func:`colour.RGB_colourspace_volume_MonteCarlo`
-   :func:`colour.RGB_colourspace_volume_coverage_MonteCarlo`
-   :func:`colour.RGB_colourspace_pointer_gamut_coverage_MonteCarlo`
-   :func:`colour.RGB_colourspace_visible_spectrum_coverage_MonteCarlo`
"""

from __future__ import annotations

import itertools
import multiprocessing
import numpy as np

from colour.algebra import random_triplet_generator
from colour.colorimetry import CCS_ILLUMINANTS
from colour.constants import DEFAULT_INT_DTYPE
from colour.hints import (
    ArrayLike,
    Callable,
    Floating,
    Integer,
    Literal,
    NDArray,
    Tuple,
    Union,
)
from colour.models import (
    Lab_to_XYZ,
    RGB_Colourspace,
    RGB_to_XYZ,
    XYZ_to_Lab,
    XYZ_to_RGB,
)
from colour.volume import is_within_pointer_gamut, is_within_visible_spectrum
from colour.utilities import as_float_array, multiprocessing_pool

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "sample_RGB_colourspace_volume_MonteCarlo",
    "RGB_colourspace_limits",
    "RGB_colourspace_volume_MonteCarlo",
    "RGB_colourspace_volume_coverage_MonteCarlo",
    "RGB_colourspace_pointer_gamut_coverage_MonteCarlo",
    "RGB_colourspace_visible_spectrum_coverage_MonteCarlo",
]


def _wrapper_RGB_colourspace_volume_MonteCarlo(arguments: Tuple) -> Integer:
    """
    Call the :func:`colour.volume.rgb.sample_RGB_colourspace_volume_MonteCarlo`
    definition with multiple arguments.

    Parameters
    ----------
    arguments
        Arguments.

    Returns
    -------
    :class:`numpy.integer`
        Inside *RGB* colourspace volume samples count.
    """

    return sample_RGB_colourspace_volume_MonteCarlo(*arguments)


def sample_RGB_colourspace_volume_MonteCarlo(
    colourspace: RGB_Colourspace,
    samples: Integer = 1000000,
    limits: ArrayLike = np.array([[0, 100], [-150, 150], [-150, 150]]),
    illuminant_Lab: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    random_generator: Callable = random_triplet_generator,
    random_state: np.random.RandomState = None,
) -> Integer:
    """
    Randomly sample the *CIE L\\*a\\*b\\** colourspace volume and returns the
    ratio of samples within the given *RGB* colourspace volume.

    Parameters
    ----------
    colourspace
        *RGB* colourspace to compute the volume of.
    samples
        Samples count.
    limits
        *CIE L\\*a\\*b\\** colourspace volume.
    illuminant_Lab
        *CIE L\\*a\\*b\\** colourspace *illuminant* chromaticity coordinates.
    chromatic_adaptation_transform
        *Chromatic adaptation* transform.
    random_generator
        Random triplet generator providing the random samples within the
        *CIE L\\*a\\*b\\** colourspace volume.
    random_state
        Mersenne Twister pseudo-random number generator to use in the random
        number generator.

    Returns
    -------
    :class:`numpy.integer`
        Within *RGB* colourspace volume samples count.

    Notes
    -----
    -   The doctest is assuming that :func:`np.random.RandomState` definition
        will return the same sequence no matter which *OS* or *Python*
        version is used. There is however no formal promise about the *prng*
        sequence reproducibility of either *Python* or *Numpy*
        implementations: Laurent. (2012). Reproducibility of python
        pseudo-random numbers across systems and versions? Retrieved January
        20, 2015, from http://stackoverflow.com/questions/8786084/\
reproducibility-of-python-pseudo-random-numbers-across-systems-and-versions

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> prng = np.random.RandomState(2)
    >>> sample_RGB_colourspace_volume_MonteCarlo(sRGB, 10e3, random_state=prng)
    ... # doctest: +ELLIPSIS
    9...
    """

    random_state = (
        random_state if random_state is not None else np.random.RandomState()
    )

    Lab = random_generator(DEFAULT_INT_DTYPE(samples), limits, random_state)
    RGB = XYZ_to_RGB(
        Lab_to_XYZ(Lab, illuminant_Lab),
        illuminant_Lab,
        colourspace.whitepoint,
        colourspace.matrix_XYZ_to_RGB,
        chromatic_adaptation_transform=chromatic_adaptation_transform,
    )
    RGB_w = RGB[
        np.logical_and(np.min(RGB, axis=-1) >= 0, np.max(RGB, axis=-1) <= 1)
    ]
    return len(RGB_w)


def RGB_colourspace_limits(colourspace: RGB_Colourspace) -> NDArray:
    """
    Compute given *RGB* colourspace volume limits in *CIE L\\*a\\*b\\**
    colourspace.

    Parameters
    ----------
    colourspace
        *RGB* colourspace to compute the volume of.

    Returns
    -------
    :class:`numpy.ndarray`
        *RGB* colourspace volume limits.

    Notes
    -----
    The limits are computed for the given *RGB* colourspace illuminant. This is
    important to account for, if the intent is to compare various *RGB*
    colourspaces together. In this instance, they must be chromatically adapted
    to the same illuminant before-hand.
    See :meth:`colour.RGB_Colourspace.chromatically_adapt` method for more
    information.

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> RGB_colourspace_limits(sRGB)  # doctest: +ELLIPSIS
    array([[   0.       ...,  100.       ...],
           [ -86.182855 ...,   98.2563272...],
           [-107.8503557...,   94.4894974...]])
    """

    Lab_c = []
    for combination in list(itertools.product([0, 1], repeat=3)):
        Lab_c.append(
            XYZ_to_Lab(
                RGB_to_XYZ(
                    combination,
                    colourspace.whitepoint,
                    colourspace.whitepoint,
                    colourspace.matrix_RGB_to_XYZ,
                ),
                colourspace.whitepoint,
            )
        )
    Lab = np.array(Lab_c)

    limits = []
    for i in np.arange(3):
        limits.append((np.min(Lab[..., i]), np.max(Lab[..., i])))

    return np.array(limits)


def RGB_colourspace_volume_MonteCarlo(
    colourspace: RGB_Colourspace,
    samples: Integer = 1000000,
    limits: ArrayLike = np.array([[0, 100], [-150, 150], [-150, 150]]),
    illuminant_Lab: ArrayLike = CCS_ILLUMINANTS[
        "CIE 1931 2 Degree Standard Observer"
    ]["D65"],
    chromatic_adaptation_transform: Union[
        Literal[
            "Bianco 2010",
            "Bianco PC 2010",
            "Bradford",
            "CAT02 Brill 2008",
            "CAT02",
            "CAT16",
            "CMCCAT2000",
            "CMCCAT97",
            "Fairchild",
            "Sharp",
            "Von Kries",
            "XYZ Scaling",
        ],
        str,
    ] = "CAT02",
    random_generator: Callable = random_triplet_generator,
    random_state: np.random.RandomState = None,
) -> Floating:
    """
    Perform given *RGB* colourspace volume computation using *Monte Carlo*
    method and multiprocessing.

    Parameters
    ----------
    colourspace\
        *RGB* colourspace to compute the volume of.
    samples\
        Samples count.
    limits\
        *CIE L\\*a\\*b\\** colourspace volume.
    illuminant_Lab\
        *CIE L\\*a\\*b\\** colourspace *illuminant* chromaticity coordinates.
    chromatic_adaptation_transform\
        *Chromatic adaptation* method.
    random_generator\
        Random triplet generator providing the random samples within the
        *CIE L\\*a\\*b\\** colourspace volume.
    random_state\
        Mersenne Twister pseudo-random number generator to use in the random
        number generator.

    Returns
    -------
    :class:`numpy.floating`
        *RGB* colourspace volume.

    Notes
    -----
    -   The doctest is assuming that :func:`np.random.RandomState` definition
        will return the same sequence no matter which *OS* or *Python*
        version is used. There is however no formal promise about the *prng*
        sequence reproducibility of either *Python* or *Numpy*
        implementations: Laurent. (2012). Reproducibility of python
        pseudo-random numbers across systems and versions? Retrieved January
        20, 2015, from http://stackoverflow.com/questions/8786084/\
reproducibility-of-python-pseudo-random-numbers-across-systems-and-versions

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> from colour.utilities import disable_multiprocessing
    >>> prng = np.random.RandomState(2)
    >>> with disable_multiprocessing():
    ...     RGB_colourspace_volume_MonteCarlo(sRGB, 10e3, random_state=prng)
    ... # doctest: +SKIP
    8...
    """

    processes = multiprocessing.cpu_count()
    process_samples = DEFAULT_INT_DTYPE(np.round(samples / processes))

    arguments = (
        colourspace,
        process_samples,
        limits,
        illuminant_Lab,
        chromatic_adaptation_transform,
        random_generator,
        random_state,
    )

    with multiprocessing_pool() as pool:
        results = pool.map(
            _wrapper_RGB_colourspace_volume_MonteCarlo,
            [arguments for _ in range(processes)],
        )

    Lab_volume = np.product(
        [np.sum(np.abs(x)) for x in as_float_array(limits)]
    )

    return Lab_volume * np.sum(results) / (process_samples * processes)


def RGB_colourspace_volume_coverage_MonteCarlo(
    colourspace: RGB_Colourspace,
    coverage_sampler: Callable,
    samples: Integer = 1000000,
    random_generator: Callable = random_triplet_generator,
    random_state: np.random.RandomState = None,
) -> Floating:
    """
    Return given *RGB* colourspace percentage coverage of an arbitrary volume.

    Parameters
    ----------
    colourspace
        *RGB* colourspace to compute the volume coverage percentage.
    coverage_sampler
        Python object responsible for checking the volume coverage.
    samples
        Samples count.
    random_generator
        Random triplet generator providing the random samples.
    random_state
        Mersenne Twister pseudo-random number generator to use in the random
        number generator.

    Returns
    -------
    :class:`numpy.floating`
        Percentage coverage of volume.

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> prng = np.random.RandomState(2)
    >>> RGB_colourspace_volume_coverage_MonteCarlo(
    ...     sRGB, is_within_pointer_gamut, 10e3, random_state=prng)
    ... # doctest: +ELLIPSIS
    81...
    """

    random_state = (
        random_state if random_state is not None else np.random.RandomState()
    )

    XYZ = random_generator(
        DEFAULT_INT_DTYPE(samples), random_state=random_state
    )
    XYZ_vs = XYZ[coverage_sampler(XYZ)]

    RGB = XYZ_to_RGB(
        XYZ_vs,
        colourspace.whitepoint,
        colourspace.whitepoint,
        colourspace.matrix_XYZ_to_RGB,
    )

    RGB_c = RGB[
        np.logical_and(np.min(RGB, axis=-1) >= 0, np.max(RGB, axis=-1) <= 1)
    ]

    return 100 * RGB_c.size / XYZ_vs.size


def RGB_colourspace_pointer_gamut_coverage_MonteCarlo(
    colourspace: RGB_Colourspace,
    samples: Integer = 1000000,
    random_generator: Callable = random_triplet_generator,
    random_state: np.random.RandomState = None,
) -> Floating:
    """
    Return given *RGB* colourspace percentage coverage of Pointer's Gamut
    volume using *Monte Carlo* method.

    Parameters
    ----------
    colourspace
        *RGB* colourspace to compute the *Pointer's Gamut* coverage percentage.
    samples
        Samples count.
    random_generator
        Random triplet generator providing the random samples.
    random_state
        Mersenne Twister pseudo-random number generator to use in the random
        number generator.

    Returns
    -------
    :class:`numpy.floating`
        Percentage coverage of *Pointer's Gamut* volume.

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> prng = np.random.RandomState(2)
    >>> RGB_colourspace_pointer_gamut_coverage_MonteCarlo(
    ...     sRGB, 10e3, random_state=prng)  # doctest: +ELLIPSIS
    81...
    """

    return RGB_colourspace_volume_coverage_MonteCarlo(
        colourspace,
        is_within_pointer_gamut,
        samples,
        random_generator,
        random_state,
    )


def RGB_colourspace_visible_spectrum_coverage_MonteCarlo(
    colourspace: RGB_Colourspace,
    samples: Integer = 1000000,
    random_generator: Callable = random_triplet_generator,
    random_state: np.random.RandomState = None,
) -> Floating:
    """
    Return given *RGB* colourspace percentage coverage of visible spectrum
    volume using *Monte Carlo* method.

    Parameters
    ----------
    colourspace
        *RGB* colourspace to compute the visible spectrum coverage percentage.
    samples
        Samples count.
    random_generator
        Random triplet generator providing the random samples.
    random_state
        Mersenne Twister pseudo-random number generator to use in the random
        number generator.

    Returns
    -------
    :class:`numpy.floating`
        Percentage coverage of visible spectrum volume.

    Examples
    --------
    >>> from colour.models import RGB_COLOURSPACE_sRGB as sRGB
    >>> prng = np.random.RandomState(2)
    >>> RGB_colourspace_visible_spectrum_coverage_MonteCarlo(
    ...     sRGB, 10e3, random_state=prng)  # doctest: +ELLIPSIS
    46...
    """

    return RGB_colourspace_volume_coverage_MonteCarlo(
        colourspace,
        is_within_visible_spectrum,
        samples,
        random_generator,
        random_state,
    )



"""
RED Colourspaces
================

Defines the *RED* colourspaces:

-   :attr:`colour.models.RGB_COLOURSPACE_RED_COLOR`
-   :attr:`colour.models.RGB_COLOURSPACE_RED_COLOR_2`
-   :attr:`colour.models.RGB_COLOURSPACE_RED_COLOR_3`
-   :attr:`colour.models.RGB_COLOURSPACE_RED_COLOR_4`
-   :attr:`colour.models.RGB_COLOURSPACE_DRAGON_COLOR`
-   :attr:`colour.models.RGB_COLOURSPACE_DRAGON_COLOR_2`
-   :attr:`colour.models.RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB`

References
----------
-   :cite:`Mansencal2015d` : Mansencal, T. (2015). RED Colourspaces Derivation.
    Retrieved May 20, 2015, from
    https://www.colour-science.org/posts/red-colourspaces-derivation
-   :cite:`Nattress2016a` : Nattress, G. (2016). Private Discussion with Shaw,
    N.
-   :cite:`SonyImageworks2012a` : Sony Imageworks. (2012). make.py. Retrieved
    November 27, 2014, from
    https://github.com/imageworks/OpenColorIO-Configs/blob/master/\
nuke-default/make.py
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    normalised_primary_matrix,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
    log_encoding_Log3G10,
    log_decoding_Log3G10,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_RED_COLOR",
    "WHITEPOINT_NAME_RED_COLOR",
    "CCS_WHITEPOINT_RED_COLOR",
    "MATRIX_RED_COLOR_TO_XYZ",
    "MATRIX_XYZ_TO_RED_COLOR",
    "RGB_COLOURSPACE_RED_COLOR",
    "PRIMARIES_RED_COLOR_2",
    "WHITEPOINT_NAME_RED_COLOR_2",
    "CCS_WHITEPOINT_RED_COLOR_2",
    "MATRIX_RED_COLOR_2_TO_XYZ",
    "MATRIX_XYZ_TO_RED_COLOR_2",
    "RGB_COLOURSPACE_RED_COLOR_2",
    "PRIMARIES_RED_COLOR_3",
    "WHITEPOINT_NAME_RED_COLOR_3",
    "CCS_WHITEPOINT_RED_COLOR_3",
    "MATRIX_RED_COLOR_3_TO_XYZ",
    "MATRIX_XYZ_TO_RED_COLOR_3",
    "RGB_COLOURSPACE_RED_COLOR_3",
    "PRIMARIES_RED_COLOR_4",
    "WHITEPOINT_NAME_RED_COLOR_4",
    "CCS_WHITEPOINT_RED_COLOR_4",
    "MATRIX_RED_COLOR_4_TO_XYZ",
    "MATRIX_XYZ_TO_RED_COLOR_4",
    "RGB_COLOURSPACE_RED_COLOR_4",
    "PRIMARIES_DRAGON_COLOR",
    "WHITEPOINT_NAME_DRAGON_COLOR",
    "CCS_WHITEPOINT_DRAGON_COLOR",
    "MATRIX_DRAGON_COLOR_TO_XYZ",
    "MATRIX_XYZ_TO_DRAGON_COLOR",
    "RGB_COLOURSPACE_DRAGON_COLOR",
    "PRIMARIES_DRAGON_COLOR_2",
    "WHITEPOINT_NAME_DRAGON_COLOR_2",
    "CCS_WHITEPOINT_DRAGON_COLOR_2",
    "MATRIX_DRAGON_COLOR_2_TO_XYZ",
    "MATRIX_XYZ_TO_DRAGON_COLOR_2",
    "RGB_COLOURSPACE_DRAGON_COLOR_2",
    "PRIMARIES_RED_WIDE_GAMUT_RGB",
    "WHITEPOINT_NAME_RED_WIDE_GAMUT_RGB",
    "CCS_WHITEPOINT_RED_WIDE_GAMUT_RGB",
    "MATRIX_RED_WIDE_GAMUT_RGB_TO_XYZ",
    "MATRIX_XYZ_TO_RED_WIDE_GAMUT_RGB",
    "RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB",
]

PRIMARIES_RED_COLOR: NDArray = np.array(
    [
        [0.701058563171395, 0.330180975940326],
        [0.298811317306316, 0.625169245953133],
        [0.135038675201355, 0.035261776551191],
    ]
)
"""*REDcolor* colourspace primaries."""

WHITEPOINT_NAME_RED_COLOR: str = "D65"
"""*REDcolor* colourspace whitepoint name."""

CCS_WHITEPOINT_RED_COLOR: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_RED_COLOR]
"""*REDcolor* colourspace whitepoint chromaticity coordinates."""

MATRIX_RED_COLOR_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_RED_COLOR, CCS_WHITEPOINT_RED_COLOR
)
"""*REDcolor* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RED_COLOR: NDArray = np.linalg.inv(MATRIX_RED_COLOR_TO_XYZ)
"""*CIE XYZ* tristimulus values to *REDcolor* colourspace matrix."""

RGB_COLOURSPACE_RED_COLOR: RGB_Colourspace = RGB_Colourspace(
    "REDcolor",
    PRIMARIES_RED_COLOR,
    CCS_WHITEPOINT_RED_COLOR,
    WHITEPOINT_NAME_RED_COLOR,
    MATRIX_RED_COLOR_TO_XYZ,
    MATRIX_XYZ_TO_RED_COLOR,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_RED_COLOR.__doc__ = """
*REDcolor* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_RED_COLOR_2: NDArray = np.array(
    [
        [0.897407221929776, 0.330776225980398],
        [0.296022094516625, 0.684635550900945],
        [0.099799512883393, -0.023000513177992],
    ]
)
"""*REDcolor2* colourspace primaries."""

WHITEPOINT_NAME_RED_COLOR_2: str = WHITEPOINT_NAME_RED_COLOR
"""*REDcolor2* colourspace whitepoint name."""

CCS_WHITEPOINT_RED_COLOR_2: NDArray = CCS_WHITEPOINT_RED_COLOR
"""*REDcolor2* colourspace whitepoint chromaticity coordinates."""

MATRIX_RED_COLOR_2_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_RED_COLOR_2, CCS_WHITEPOINT_RED_COLOR_2
)
"""*REDcolor2* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RED_COLOR_2: NDArray = np.linalg.inv(MATRIX_RED_COLOR_2_TO_XYZ)
"""*CIE XYZ* tristimulus values to *REDcolor2* colourspace matrix."""

RGB_COLOURSPACE_RED_COLOR_2: RGB_Colourspace = RGB_Colourspace(
    "REDcolor2",
    PRIMARIES_RED_COLOR_2,
    CCS_WHITEPOINT_RED_COLOR_2,
    WHITEPOINT_NAME_RED_COLOR_2,
    MATRIX_RED_COLOR_2_TO_XYZ,
    MATRIX_XYZ_TO_RED_COLOR_2,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_RED_COLOR_2.__doc__ = """
*REDcolor2* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_RED_COLOR_3: NDArray = np.array(
    [
        [0.702598658589917, 0.330185588938484],
        [0.295782235737268, 0.689748258397534],
        [0.111090529079787, -0.004332320984771],
    ]
)
"""*REDcolor3* colourspace primaries."""

WHITEPOINT_NAME_RED_COLOR_3: str = WHITEPOINT_NAME_RED_COLOR
"""*REDcolor3* colourspace whitepoint name."""

CCS_WHITEPOINT_RED_COLOR_3: NDArray = CCS_WHITEPOINT_RED_COLOR
"""*REDcolor3* colourspace whitepoint chromaticity coordinates."""

MATRIX_RED_COLOR_3_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_RED_COLOR_3, CCS_WHITEPOINT_RED_COLOR_3
)
"""*REDcolor3* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RED_COLOR_3: NDArray = np.linalg.inv(MATRIX_RED_COLOR_3_TO_XYZ)
"""*CIE XYZ* tristimulus values to *REDcolor3* colourspace matrix."""

RGB_COLOURSPACE_RED_COLOR_3: RGB_Colourspace = RGB_Colourspace(
    "REDcolor3",
    PRIMARIES_RED_COLOR_3,
    CCS_WHITEPOINT_RED_COLOR_3,
    WHITEPOINT_NAME_RED_COLOR_3,
    MATRIX_RED_COLOR_3_TO_XYZ,
    MATRIX_XYZ_TO_RED_COLOR_3,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_RED_COLOR_3.__doc__ = """
*REDcolor3* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_RED_COLOR_4: NDArray = np.array(
    [
        [0.702598154635438, 0.330185096210515],
        [0.295782328047083, 0.689748253964859],
        [0.144459236489795, 0.050837720977386],
    ]
)
"""*REDcolor4* colourspace primaries."""

WHITEPOINT_NAME_RED_COLOR_4: str = WHITEPOINT_NAME_RED_COLOR
"""*REDcolor4* colourspace whitepoint name."""

CCS_WHITEPOINT_RED_COLOR_4: NDArray = CCS_WHITEPOINT_RED_COLOR
"""*REDcolor4* colourspace whitepoint chromaticity coordinates."""

MATRIX_RED_COLOR_4_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_RED_COLOR_4, CCS_WHITEPOINT_RED_COLOR_4
)
"""*REDcolor4* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RED_COLOR_4: NDArray = np.linalg.inv(MATRIX_RED_COLOR_4_TO_XYZ)
"""*CIE XYZ* tristimulus values to *REDcolor4* colourspace matrix."""

RGB_COLOURSPACE_RED_COLOR_4: RGB_Colourspace = RGB_Colourspace(
    "REDcolor4",
    PRIMARIES_RED_COLOR_4,
    CCS_WHITEPOINT_RED_COLOR_4,
    WHITEPOINT_NAME_RED_COLOR_4,
    MATRIX_RED_COLOR_4_TO_XYZ,
    MATRIX_XYZ_TO_RED_COLOR_4,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_RED_COLOR_4.__doc__ = """
*REDcolor4* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_DRAGON_COLOR: NDArray = np.array(
    [
        [0.758655892599321, 0.330355348611293],
        [0.294923619810175, 0.708053242065117],
        [0.085961601167585, -0.045879436983969],
    ]
)
"""*DRAGONcolor* colourspace primaries."""

WHITEPOINT_NAME_DRAGON_COLOR: str = WHITEPOINT_NAME_RED_COLOR
"""*DRAGONcolor* colourspace whitepoint name."""

CCS_WHITEPOINT_DRAGON_COLOR: NDArray = CCS_WHITEPOINT_RED_COLOR
"""*DRAGONcolor* colourspace whitepoint chromaticity coordinates."""

MATRIX_DRAGON_COLOR_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_DRAGON_COLOR, CCS_WHITEPOINT_DRAGON_COLOR
)
"""*DRAGONcolor* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_DRAGON_COLOR: NDArray = np.linalg.inv(MATRIX_DRAGON_COLOR_TO_XYZ)
"""*CIE XYZ* tristimulus values to *DRAGONcolor* colourspace matrix."""

RGB_COLOURSPACE_DRAGON_COLOR: RGB_Colourspace = RGB_Colourspace(
    "DRAGONcolor",
    PRIMARIES_DRAGON_COLOR,
    CCS_WHITEPOINT_DRAGON_COLOR,
    WHITEPOINT_NAME_DRAGON_COLOR,
    MATRIX_DRAGON_COLOR_TO_XYZ,
    MATRIX_XYZ_TO_DRAGON_COLOR,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_DRAGON_COLOR.__doc__ = """
*DRAGONcolor* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_DRAGON_COLOR_2: NDArray = np.array(
    [
        [0.758656214177604, 0.330355835762678],
        [0.294923887732982, 0.708053363192126],
        [0.144168726866337, 0.050357384587121],
    ]
)
"""*DRAGONcolor2* colourspace primaries."""

WHITEPOINT_NAME_DRAGON_COLOR_2: str = WHITEPOINT_NAME_RED_COLOR
"""*DRAGONcolor2* colourspace whitepoint name."""

CCS_WHITEPOINT_DRAGON_COLOR_2: NDArray = CCS_WHITEPOINT_RED_COLOR
"""*DRAGONcolor2* colourspace whitepoint chromaticity coordinates."""

MATRIX_DRAGON_COLOR_2_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_DRAGON_COLOR_2, CCS_WHITEPOINT_DRAGON_COLOR_2
)
"""*DRAGONcolor2* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_DRAGON_COLOR_2: NDArray = np.linalg.inv(
    MATRIX_DRAGON_COLOR_2_TO_XYZ
)
"""*CIE XYZ* tristimulus values to *DRAGONcolor2* colourspace matrix."""

RGB_COLOURSPACE_DRAGON_COLOR_2: RGB_Colourspace = RGB_Colourspace(
    "DRAGONcolor2",
    PRIMARIES_DRAGON_COLOR_2,
    CCS_WHITEPOINT_DRAGON_COLOR_2,
    WHITEPOINT_NAME_DRAGON_COLOR_2,
    MATRIX_DRAGON_COLOR_2_TO_XYZ,
    MATRIX_XYZ_TO_DRAGON_COLOR_2,
    log_encoding_REDLogFilm,
    log_decoding_REDLogFilm,
)
RGB_COLOURSPACE_DRAGON_COLOR_2.__doc__ = """
*DRAGONcolor2* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`SonyImageworks2012a`
"""

PRIMARIES_RED_WIDE_GAMUT_RGB: NDArray = np.array(
    [
        [0.780308, 0.304253],
        [0.121595, 1.493994],
        [0.095612, -0.084589],
    ]
)
"""*REDWideGamutRGB* colourspace primaries."""

WHITEPOINT_NAME_RED_WIDE_GAMUT_RGB: str = WHITEPOINT_NAME_RED_COLOR
"""*REDWideGamutRGB* colourspace whitepoint name."""

CCS_WHITEPOINT_RED_WIDE_GAMUT_RGB = CCS_WHITEPOINT_RED_COLOR
"""*REDWideGamutRGB* colourspace whitepoint chromaticity coordinates."""

MATRIX_RED_WIDE_GAMUT_RGB_TO_XYZ: NDArray = np.array(
    [
        [0.735275, 0.068609, 0.146571],
        [0.286694, 0.842979, -0.129673],
        [-0.079681, -0.347343, 1.516082],
    ]
)
"""*REDWideGamutRGB* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_RED_WIDE_GAMUT_RGB: NDArray = np.linalg.inv(
    MATRIX_RED_WIDE_GAMUT_RGB_TO_XYZ
)
"""*CIE XYZ* tristimulus values to *REDWideGamutRGB* colourspace matrix."""

RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB: RGB_Colourspace = RGB_Colourspace(
    "REDWideGamutRGB",
    PRIMARIES_RED_WIDE_GAMUT_RGB,
    CCS_WHITEPOINT_RED_WIDE_GAMUT_RGB,
    WHITEPOINT_NAME_RED_WIDE_GAMUT_RGB,
    MATRIX_RED_WIDE_GAMUT_RGB_TO_XYZ,
    MATRIX_XYZ_TO_RED_WIDE_GAMUT_RGB,
    log_encoding_Log3G10,
    log_decoding_Log3G10,
)
RGB_COLOURSPACE_RED_WIDE_GAMUT_RGB.__doc__ = """
*REDWideGamutRGB* colourspace.

References
----------
:cite:`Mansencal2015d`, :cite:`Nattress2016a`, :cite:`SonyImageworks2012a`
"""
